@@ -24,12 +24,14 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
+#include <string.h>
 #include <glib/gerror.h>
 #include <glib/gtypes.h>
+#include <glib/gi18n.h>
 #include <glib-object.h>
 #include <gst/gst.h>
 #include <gst/gconf/gconf.h>
-#include <libgnome/gnome-i18n.h>
+#include <gst/tag/tag.h>
 #include <profiles/gnome-media-profiles.h>
 #include "sj-extractor.h"
 #include "sj-structures.h"
@@ -50,24 +52,28 @@ enum {
   LAST_SIGNAL
 };
 
-static int sje_table_signals[LAST_SIGNAL] = { 0 };
+static guint sje_table_signals[LAST_SIGNAL] = { 0 };
 
 #define DEFAULT_AUDIO_PROFILE_NAME "cdlossy"
 
 struct SjExtractorPrivate {
+  /** The current audio profile */
   GMAudioProfile *profile;
+  /** If the pipeline needs to be re-created */
   gboolean rebuild_pipeline;
-  GstElement *pipeline;
-  GstElement *cdparanoia, *encoder, *filesink;
+  /* The gstreamer pipeline elements */
+  GstElement *pipeline, *cdparanoia, *queue, *thread, *encoder, *filesink;
   GstFormat track_format;
   GstPad *source_pad;
-  const char *encoder_name;
   char *device_path;
   int paranoia_mode;
   int track_start;
   int seconds;
   GError *construct_error;
+  /* Variables used by the thread to push signals into the main thread */
   guint idle_id;
+  GMutex *idle_mutex;
+  GAsyncQueue *idle_queue;
 };
 
 static void build_pipeline (SjExtractor *extractor);
@@ -76,39 +82,13 @@ static void build_pipeline (SjExtractor *extractor);
  * GObject methods
  */
 
-static void sj_extractor_class_init (SjExtractorClass *klass);
-static void sj_extractor_instance_init (SjExtractor *bcs);
 static void sj_extractor_set_property (GObject *object, guint property_id,
                                        const GValue *value, GParamSpec *pspec);
 static void sj_extractor_get_property (GObject *object, guint property_id,
                                        GValue *value, GParamSpec *pspec);
 static void sj_extractor_finalize (GObject *object);
 
-
-GType sj_extractor_get_type (void)
-{
-  static GType sj_extractor_type = 0;
-  
-  if (!sj_extractor_type) {
-    static const GTypeInfo sj_extractor_info = {
-      sizeof (SjExtractorClass),
-      (GBaseInitFunc) NULL,
-      (GBaseFinalizeFunc) NULL,
-      (GClassInitFunc) sj_extractor_class_init,
-      (GClassFinalizeFunc) NULL,
-      NULL /* class_data */,
-      sizeof (SjExtractor),
-      0 /* n_preallocs */,
-      (GInstanceInitFunc) sj_extractor_instance_init,
-      NULL
-    };
-    
-    sj_extractor_type = g_type_register_static (G_TYPE_OBJECT,
-                                                "SjExtractor", &sj_extractor_info,
-                                                (GTypeFlags)0);
-  }
-  return sj_extractor_type;
-}
+G_DEFINE_TYPE(SjExtractor, sj_extractor, G_TYPE_OBJECT);
 
 static void sj_extractor_class_init (SjExtractorClass *klass)
 {
@@ -152,11 +132,14 @@ static void sj_extractor_class_init (SjExtractorClass *klass)
 }
 
 
-static void sj_extractor_instance_init (SjExtractor *extractor)
+static void sj_extractor_init (SjExtractor *extractor)
 {
   extractor->priv = g_new0 (SjExtractorPrivate, 1);
   extractor->priv->profile = gm_audio_profile_lookup(DEFAULT_AUDIO_PROFILE_NAME);
   extractor->priv->rebuild_pipeline = TRUE;
+  extractor->priv->idle_id = 0;
+  extractor->priv->idle_mutex = g_mutex_new ();
+  extractor->priv->idle_queue = g_async_queue_new ();
 }
 
 static void sj_extractor_set_property (GObject *object, guint property_id,
@@ -189,37 +172,77 @@ static void sj_extractor_get_property (GObject *object, guint property_id,
 static void sj_extractor_finalize (GObject *object)
 {
   SjExtractor *extractor = (SjExtractor *)object;
-  if (extractor->priv->idle_id > 0)
-    g_source_remove (extractor->priv->idle_id);
   if (extractor->priv->pipeline)
     gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
   /* TODO: free the members of priv */
+  g_mutex_free (extractor->priv->idle_mutex);
+  g_async_queue_unref (extractor->priv->idle_queue);
   g_free (extractor->priv);
-  /* TODO: SJ_EXTRACTOR_GET_CLASS (extractor)->parent_class->finalize (object);*/
-  /* g_object_finalize (object); */
   extractor->priv = NULL;
-  extractor = NULL;
+  G_OBJECT_CLASS (sj_extractor_parent_class)->finalize (object);
 }
 
 /*
  * Private Methods
  */
+
+typedef struct {
+  guint signal;
+  union {
+    GError *error;
+    int progress;
+  } args;
+} SignalData;
+
+static gboolean
+fire_signal(gpointer user_data)
+{
+  SjExtractor *extractor = SJ_EXTRACTOR (user_data);
+  SignalData *data;
+  g_mutex_lock (extractor->priv->idle_mutex);
+
+  while ((data = g_async_queue_try_pop (extractor->priv->idle_queue)) != NULL) {
+    switch (data->signal) {
+    case COMPLETION:
+      g_signal_emit (extractor, sje_table_signals [COMPLETION], 0);
+      break;
+    case ERROR:
+      g_signal_emit (extractor, sje_table_signals[ERROR], 0, data->args.error);
+      g_error_free (data->args.error);
+      break;
+    case PROGRESS:
+      g_signal_emit (extractor, sje_table_signals[PROGRESS], 0, data->args.progress);
+      break;
+    default:
+      g_warning ("Unknown signal ID %d", data->signal);
+    }
+    g_free (data);
+  }
+
+  extractor->priv->idle_id = 0;
+  g_mutex_unlock (extractor->priv->idle_mutex);
+  return FALSE;
+}
+
+static void
+queue_signal (SjExtractor *extractor, SignalData *data)
+{
+  g_mutex_lock (extractor->priv->idle_mutex);
+  g_async_queue_push (extractor->priv->idle_queue, data);
+  if (extractor->priv->idle_id == 0)
+    extractor->priv->idle_id = g_idle_add (fire_signal, extractor);
+  g_mutex_unlock (extractor->priv->idle_mutex);
+}
+
 static void eos_cb (GstElement *gstelement, SjExtractor *extractor)
 {
+  SignalData *data;
   g_return_if_fail (SJ_IS_EXTRACTOR (extractor));
   gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
   extractor->priv->rebuild_pipeline = TRUE;
-  g_signal_emit (G_OBJECT (extractor),
-                 sje_table_signals[COMPLETION],
-                 0);
-}
-
-static const char* get_encoder_name (SjExtractor *extractor)
-{
-  SjExtractorPrivate *priv;
-  g_return_val_if_fail (SJ_IS_EXTRACTOR (extractor), NULL);
-  priv = (SjExtractorPrivate*)extractor->priv;
-  return priv->encoder_name;
+  data = g_new0 (SignalData, 1);
+  data->signal = COMPLETION;
+  queue_signal (extractor, data);
 }
 
 static GstElement* build_encoder (SjExtractor *extractor)
@@ -240,11 +263,23 @@ static GstElement* build_encoder (SjExtractor *extractor)
 static void error_cb (GstElement *element, GObject *arg, GError *err, gchar *arg2, gpointer user_data)
 {
   SjExtractor *extractor = SJ_EXTRACTOR (user_data);
+  SignalData *data;
   /* Make sure the pipeline is not running any more */
   gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
-  g_signal_emit (G_OBJECT (extractor),
-                 sje_table_signals[ERROR],
-                 0, err);
+  data = g_new0 (SignalData, 1);
+  data->signal = ERROR;
+  data->args.error = g_error_copy (err);
+  queue_signal (extractor, data);
+}
+
+/**
+ * Callback from the gnomevfssink to say that its about to overwrite a file.
+ * For now, Just Say Yes.  If this API will stay in 0.9, then rewrite
+ * SjExtractor.
+ */
+static gboolean just_say_yes (GstElement *element, gpointer filename, gpointer user_data)
+{
+  return TRUE;
 }
 
 static void build_pipeline (SjExtractor *extractor)
@@ -259,7 +294,7 @@ static void build_pipeline (SjExtractor *extractor)
   if (priv->pipeline != NULL) {
     gst_object_unref (GST_OBJECT (priv->pipeline));
   }
-  priv->pipeline = gst_pipeline_new ("pipeline");
+  priv->pipeline = gst_thread_new ("pipeline");
   g_signal_connect (G_OBJECT (priv->pipeline), "error", G_CALLBACK (error_cb), extractor);
 
   /* Read from CD */
@@ -281,33 +316,47 @@ static void build_pipeline (SjExtractor *extractor)
   priv->source_pad = gst_element_get_pad (priv->cdparanoia, "src");
   g_assert (priv->source_pad); /* TODO: GError */
 
+  priv->queue = gst_element_factory_make ("queue", "queue");
+  /* Nice big buffers... */
+  g_object_set (priv->queue, "max-size-time", 20 * GST_SECOND, NULL);
+  
+  priv->thread = gst_thread_new ("encode thread");
+
   /* Encode */
   priv->encoder = build_encoder (extractor);
   if (priv->encoder == NULL) {
     g_set_error (&priv->construct_error,
                  SJ_ERROR, SJ_ERROR_INTERNAL_ERROR,
-                 _("Could not create GStreamer encoder (%s)"), get_encoder_name (extractor));
+                 _("Could not create GStreamer encoders for %s"), gm_audio_profile_get_name (priv->profile));
     return;
   }
   /* Connect to the eos so we know when its finished */
   g_signal_connect (priv->encoder, "eos", G_CALLBACK (eos_cb), extractor);
 
   /* Write to disk */
-  priv->filesink = gst_element_factory_make ("filesink", "filesink");
+  priv->filesink = gst_element_factory_make ("gnomevfssink", "gnomevfssink");
   if (priv->filesink == NULL) {
     g_set_error (&priv->construct_error,
                  SJ_ERROR, SJ_ERROR_INTERNAL_ERROR,
                  _("Could not create GStreamer file output"));
     return;
   }
+  g_signal_connect (G_OBJECT (priv->filesink), "allow-overwrite", G_CALLBACK (just_say_yes), extractor);
 
   /* Add the elements to the pipeline */
   gst_bin_add (GST_BIN (priv->pipeline), priv->cdparanoia);
-  gst_bin_add (GST_BIN (priv->pipeline), priv->encoder);
-  gst_bin_add (GST_BIN (priv->pipeline), priv->filesink);
+  gst_bin_add (GST_BIN (priv->thread), priv->queue);
+  gst_bin_add (GST_BIN (priv->thread), priv->encoder);
+  gst_bin_add (GST_BIN (priv->thread), priv->filesink);
+  gst_bin_add (GST_BIN (priv->pipeline), priv->thread);
 
   /* Link it all together */
-  gst_element_link_many (priv->cdparanoia, priv->encoder, priv->filesink, NULL);
+  if (!gst_element_link_many (priv->cdparanoia, priv->queue, priv->encoder, priv->filesink, NULL)) {
+    /* TODO: need to produce a GError here */
+    g_warning ("Cannot link pipeline, very bad!");
+    g_object_unref (priv->pipeline);
+    return;
+  }
 
   priv->rebuild_pipeline = FALSE;
 }
@@ -333,15 +382,14 @@ static gboolean tick_timeout_cb(SjExtractor *extractor)
 
   secs = nanos / GST_SECOND;
   if (secs != extractor->priv->seconds) {
+    SignalData *data;
     extractor->priv->seconds = secs;
-    g_signal_emit (G_OBJECT (extractor),
-                   sje_table_signals[PROGRESS],
-                   0, secs - extractor->priv->track_start);
+    data = g_new0 (SignalData, 1);
+    data->signal = PROGRESS;
+    data->args.progress = secs - extractor->priv->track_start;
+    queue_signal (extractor, data);
   }
 
-  /* Extra kick in the pants to keep things moving on a busy system */
-  gst_bin_iterate (GST_BIN (extractor->priv->pipeline));
-  
   return TRUE;
 }
 
@@ -389,7 +437,7 @@ void sj_extractor_set_paranoia (SjExtractor *extractor, const int paranoia_mode)
     g_object_set (G_OBJECT (extractor->priv->cdparanoia), "paranoia-mode", paranoia_mode, NULL);
 }
 
-void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *track, const char* path, GError **error)
+void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *track, const char* url, GError **error)
 {
   GstEvent *event;
   static GstFormat format = GST_FORMAT_TIME;
@@ -399,16 +447,13 @@ void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *tra
   g_return_if_fail (extractor != NULL);
   g_return_if_fail (SJ_IS_EXTRACTOR (extractor));
 
-  g_return_if_fail (path != NULL);
+  g_return_if_fail (url != NULL);
   g_return_if_fail (track != NULL);
 
   priv = extractor->priv;
 
   /* See if we need to rebuild the pipeline */
   if (priv->rebuild_pipeline != FALSE) {
-    if (priv->idle_id > 0)
-      g_source_remove (priv->idle_id);
-    priv->idle_id = 0;
     build_pipeline (extractor);
     if (priv->construct_error != NULL) {
       *error = g_error_copy (priv->construct_error);
@@ -420,7 +465,7 @@ void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *tra
 
   /* Set the output filename */
   gst_element_set_state (priv->filesink, GST_STATE_NULL);
-  g_object_set (G_OBJECT (priv->filesink), "location", path, NULL);
+  g_object_set (G_OBJECT (priv->filesink), "location", url, NULL);
 
   /* Set the metadata */
   {
@@ -443,12 +488,21 @@ void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *tra
                             GST_TAG_ALBUM, track->album->title,
                             GST_TAG_ENCODER, _("Sound Juicer"),
                             GST_TAG_ENCODER_VERSION, VERSION,
+                            GST_TAG_MUSICBRAINZ_ALBUMID, track->album->album_id,
+                            GST_TAG_MUSICBRAINZ_ALBUMARTISTID, track->album->artist_id,
+                            GST_TAG_MUSICBRAINZ_ARTISTID, track->artist_id,
+                            GST_TAG_MUSICBRAINZ_TRACKID, track->track_id,
                             NULL);
-        if (track->genre != LAST_GENRE) {
+        if (track->album->genre != NULL && strcmp (track->album->genre, "") != 0) {
           gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                              /* TODO: is APPEND right? */
                               GST_TAG_MERGE_APPEND,
-                              GST_TAG_GENRE, sj_genre_name (track->genre),
+                              GST_TAG_GENRE, track->album->genre,
+                              NULL);
+        }
+        if (track->album->release_date) {
+          gst_tag_setter_add (GST_TAG_SETTER (elt->data),
+                              GST_TAG_MERGE_APPEND,
+                              GST_TAG_DATE, g_date_get_julian (track->album->release_date),
                               NULL);
         }
         taggable = TRUE;
@@ -479,7 +533,7 @@ void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *tra
   priv->track_start = nanos / GST_SECOND;
 
   gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
-  priv->idle_id = g_idle_add ((GSourceFunc)gst_bin_iterate, priv->pipeline);
+  
   g_timeout_add (200, (GSourceFunc)tick_timeout_cb, extractor);
 }
 
@@ -491,16 +545,9 @@ void sj_extractor_cancel_extract (SjExtractor *extractor)
   if (gst_element_get_state (extractor->priv->pipeline) != GST_STATE_PLAYING) {
     return;
   }
-  gst_element_set_state (extractor->priv->pipeline, GST_STATE_PAUSED);
-  g_source_remove (extractor->priv->idle_id);
+  gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
 
   extractor->priv->rebuild_pipeline = TRUE;
-
-  /*
-   * TODO: I don't think I need to remove gst from the idle loop, as
-   * it will call gst_bin_iterate(), which sees that it is PAUSED,
-   * will return FALSE as it did nothing, and be removed. Correct?
-   */
 }
 
 gboolean sj_extractor_supports_encoding (GError **error)
